@@ -8,13 +8,13 @@
 --![Usage example](../images/AUTOGEN_wibox_layout_defaults_ratio.svg)
 --
 -- @usage
---local w = wibox.widget {
---    generic_widget( 'first'  ),
---    generic_widget( 'second' ),
---    generic_widget( 'third'  ),
---    layout  = wibox.layout.ratio.horizontal
---}
---w:ajust_ratio(2, 0.44, 0.33, 0.22)
+-- local w = wibox.widget {
+--     generic_widget( 'first'  ),
+--     generic_widget( 'second' ),
+--     generic_widget( 'third'  ),
+--     layout  = wibox.layout.ratio.horizontal
+-- }
+-- w:ajust_ratio(2, 0.44, 0.33, 0.22)
 -- @author Emmanuel Lepage Vallee
 -- @copyright 2016 Emmanuel Lepage Vallee
 -- @classmod wibox.layout.ratio
@@ -25,13 +25,13 @@ local flex  = require("wibox.layout.flex" )
 local table = table
 local pairs = pairs
 local floor = math.floor
-local util  = require("awful.util")
+local gmath = require("gears.math")
+local gtable = require("gears.table")
+local unpack = unpack or table.unpack -- luacheck: globals unpack (compatibility with Lua 5.1)
 
 local ratio = {}
 
---Imported documentation
-
---- Set a widget at a specific index, replace the current one.
+----- Set a widget at a specific index, replace the current one.
 -- **Signal:** widget::replaced The argument is the new widget and the old one
 -- and the index.
 -- @tparam number index A widget or a widget index
@@ -45,7 +45,7 @@ local ratio = {}
 -- and the index.
 -- @param widget The widget to replace
 -- @param widget2 The widget to replace `widget` with
--- @tparam[opt=false] boolean recursive Digg in all compatible layouts to find the widget.
+-- @tparam[opt=false] boolean recursive Dig in all compatible layouts to find the widget.
 -- @treturn boolean If the operation is successful
 -- @name replace_widget
 -- @class function
@@ -64,7 +64,7 @@ local ratio = {}
 -- if the layouts not the same, then only `widget::replaced` will be emitted.
 -- @param widget1 The first widget
 -- @param widget2 The second widget
--- @tparam[opt=false] boolean recursive Digg in all compatible layouts to find the widget.
+-- @tparam[opt=false] boolean recursive Dig in all compatible layouts to find the widget.
 -- @treturn boolean If the operation is successful
 -- @name swap_widgets
 -- @class function
@@ -79,6 +79,77 @@ local ratio = {}
 -- @name reset
 -- @class function
 
+--- The widget used to fill the spacing between the layout elements.
+--
+-- By default, no widget is used.
+--
+--
+--
+--![Usage example](../images/AUTOGEN_wibox_layout_ratio_spacing_widget.svg)
+--
+-- @usage
+-- -- Use the separator widget directly
+-- local w1 = wibox.widget {
+--     spacing        = 10,
+--     spacing_widget = wibox.widget.separator,
+--     layout         = wibox.layout.ratio.horizontal
+-- }
+-- -- Use a standard declarative widget construct
+-- local w2 = wibox.widget {
+--     spacing = 10,
+--     spacing_widget = {
+--         color  = '#00ff00',
+--         shape  = gears.shape.circle,
+--         widget = wibox.widget.separator,
+--     },
+--     layout = wibox.layout.ratio.horizontal
+-- }
+-- -- Use composed widgets
+-- local w3 = wibox.widget {
+--     spacing = 10,
+--     spacing_widget = {
+--         {
+--             text   = 'F',
+--             widget = wibox.widget.textbox,
+--         },
+--         bg     = '#ff0000',
+--         widget = wibox.container.background,
+--     },
+--     layout = wibox.layout.ratio.horizontal
+-- }
+-- -- Use negative spacing to create a powerline effect
+-- local w4 = wibox.widget {
+--     spacing = -12,
+--     spacing_widget = {
+--         color  = '#ff0000',
+--         shape  = gears.shape.powerline,
+--         widget = wibox.widget.separator,
+--     },
+--     layout = wibox.layout.ratio.horizontal
+-- }
+--
+-- @property spacing_widget
+-- @param widget
+
+--- Add spacing between each layout widgets.
+--
+--
+--
+--![Usage example](../images/AUTOGEN_wibox_layout_ratio_spacing.svg)
+--
+-- @usage
+-- for i=1, 5 do
+--     local w = wibox.widget {
+--         first,
+--         second,
+--         third,
+--         spacing = i*3,
+--         layout  = wibox.layout.ratio.horizontal
+--     }
+-- end
+--
+-- @property spacing
+-- @tparam number spacing Spacing between widgets.
 
 -- Compute the sum of all ratio (ideally, it should be 1)
 local function gen_sum(self, i_s, i_e)
@@ -137,42 +208,132 @@ local function normalize(self)
     assert(new_sum > 0.99 and new_sum < 1.01)
 end
 
-function ratio:layout(_, width, height)
-    local result = {}
+function ratio:layout(context, width, height)
+    local preliminary_results = {}
     local pos,spacing = 0, self._private.spacing
+    local strategy = self:get_inner_fill_strategy()
+    local has_stragety = strategy ~= "default"
+    local to_redistribute, void_count = 0, 0
+    local dir = self._private.dir or "x"
+    local spacing_widget = self._private.spacing_widget
+    local abspace = math.abs(spacing)
+    local spoffset = spacing < 0 and 0 or spacing
+    local is_y = self._private.dir == "y"
+    local is_x = not is_y
 
     for k, v in ipairs(self._private.widgets) do
-        local space
-        local x, y, w, h
+        local space, is_void
+        local x, y, w, h, fit_h, fit_w
 
-        if self._private.dir == "y" then
+        if dir == "y" then
             space = height * self._private.ratios[k]
-            x, y = 0, util.round(pos)
+            x, y = 0, gmath.round(pos)
             w, h = width, floor(space)
         else
             space = width * self._private.ratios[k]
-            x, y = util.round(pos), 0
+            x, y = gmath.round(pos), 0
             w, h = floor(space), height
         end
 
-        table.insert(result, base.place_widget_at(v, x, y, w, h))
+        -- Keep track of the unused entries
+        if has_stragety then
+            fit_h, fit_w = base.fit_widget(
+                self, context, v,
+                dir == "x" and floor(space) or w,
+                dir == "y" and floor(space) or h
+            )
+
+            is_void = (v.visible == false)
+                or (dir == "x" and fit_w == 0)
+                or (dir == "y" and fit_h == 0)
+
+            if is_void then
+                to_redistribute = to_redistribute + space + spacing
+                void_count = void_count + 1
+            end
+        end
+
+        table.insert(preliminary_results, {v, x, y, w, h, is_void})
 
         pos = pos + space + spacing
 
         -- Make sure all widgets fit in the layout, if they aren't, something
         -- went wrong
-        if (self._private.dir == "y" and util.round(pos) >= height) or
-            (self._private.dir ~= "y" and util.round(pos) >= width) then
+        if (dir == "y" and gmath.round(pos) >= height) or
+            (dir ~= "y" and gmath.round(pos) >= width) then
             break
         end
     end
 
+    local active = #preliminary_results - void_count
+    local result, real_pos, space_front = {}, 0, strategy == "right" and
+        to_redistribute or (
+            strategy == "center" and math.floor(to_redistribute/2) or 0
+        )
+
+    -- The number of spaces between `n` element is `n-1`, if there is spaces
+    -- outside, then it is `n+1`
+    if strategy == "spacing" then
+        space_front = (space_front+to_redistribute/(active + 1))
+        to_redistribute = (to_redistribute/(active + 1))*(active - 1)
+    end
+
+    spacing = strategy:match("spacing")
+        and to_redistribute/(active - 1) or 0
+
+    -- Only the `justify` strategy changes the original widget size.
+    to_redistribute = (strategy == "justify") and to_redistribute or 0
+
+    for k, entry in ipairs(preliminary_results) do
+        local v, x, y, w, h, is_void = unpack(entry)
+
+        -- Redistribute the space or move the widgets
+        if strategy ~= "default" then
+            if dir == "y" then
+                h = is_void and 0 or h + (to_redistribute / (active))
+                y = space_front + real_pos
+                real_pos = real_pos + h + (is_void and 0 or spacing)
+
+            else
+                w = is_void and 0 or w + (to_redistribute / (active))
+                x = space_front + real_pos
+                real_pos = real_pos + w + (is_void and 0 or spacing)
+            end
+        end
+
+        if k > 1 and abspace > 0 and spacing_widget then
+            table.insert(result, base.place_widget_at(
+                spacing_widget, is_x and (x - spoffset) or x, is_y and (y - spoffset) or y,
+                is_x and abspace or w, is_y and abspace or h
+            ))
+        end
+
+        table.insert(result, base.place_widget_at(v, x, y, w, h))
+    end
+
     return result
 end
 
 --- Increase the ratio of "widget"
 -- If the increment produce an invalid ratio (not between 0 and 1), the method
 -- do nothing.
+--
+--
+--
+--![Usage example](../images/AUTOGEN_wibox_layout_ratio_inc_ratio.svg)
+--
+-- @usage
+-- local ret = wibox.layout.fixed.vertical()
+-- local w = wibox.widget {
+--     first,
+--     second,
+--     third,
+--     layout  = wibox.layout.ratio.horizontal
+-- }
+-- for i=1, 5 do
+--     w:inc_ratio(2, 0.1)
+-- end
+--
 -- @tparam number index The widget index to change
 -- @tparam number increment An floating point value between -1 and 1 where the
 --   end result is within 0 and 1
@@ -246,7 +407,31 @@ function ratio:set_widget_ratio(widget, percent)
 end
 
 --- Update all widgets to match a set of a ratio.
--- The sum of before, itself and after must be 1 or nothing will be done
+-- The sum of before, itself and after must be 1 or nothing will be done.
+--
+--
+--
+--![Usage example](../images/AUTOGEN_wibox_layout_ratio_ajust_ratio.svg)
+--
+-- @usage
+-- local ret = wibox.layout.fixed.vertical()
+-- local w = wibox.widget {
+--     first,
+--     second,
+--     third,
+--     layout  = wibox.layout.ratio.horizontal
+-- }
+-- local values = {
+--     {0.25, 0.50, 0.25},
+--     {0.33, 0.55, 0.12},
+--     {0.123, 0.456, 0.789},
+--     {0.123, 0, 0.789},
+--     {0, 1, 0},
+-- }
+-- for i=1, 5 do
+--     w:ajust_ratio(2, unpack(values[i]))
+-- end
+--
 -- @tparam number index The index of the widget to change
 -- @tparam number before The sum of the ratio before the widget
 -- @tparam number itself The ratio for "widget"
@@ -350,10 +535,51 @@ function ratio:insert(index, widget)
     self:emit_signal("widget::inserted", widget, #self._private.widgets)
 end
 
+--- Set how the space of invisible or `0x0` sized widget is redistributed.
+--
+-- Possible values:
+--
+-- * "default": Honor the ratio and do not redistribute the space.
+-- * "justify": Distribute the space among remaining widgets.
+-- * "center": Squash remaining widgets and leave equal space on both side.
+-- * "inner_spacing": Add equal spacing between all widgets.
+-- * "spacing": Add equal spacing between all widgets and on the outside.
+-- * "left": Squash remaining widgets and leave empty space on the left.
+-- * "right": Squash remaining widgets and leave empty space on the right.
+--
+--
+--
+--![Usage example](../images/AUTOGEN_wibox_layout_ratio_strategy.svg)
+--
+--
+-- @property inner_fill_strategy
+-- @tparam string inner_fill_strategy One of the value listed above.
+
+function ratio:get_inner_fill_strategy()
+    return self._private.inner_fill_strategy or "default"
+end
+
+local valid_strategies = {
+    default       = true,
+    justify       = true,
+    center        = true,
+    inner_spacing = true,
+    spacing       = true,
+    left          = true,
+    right         = true
+}
+
+function ratio:set_inner_fill_strategy(strategy)
+    assert(valid_strategies[strategy] ~= nil, "Invalid strategy: "..(strategy or ""))
+
+    self._private.inner_fill_strategy = strategy
+    self:emit_signal("widget::layout_changed")
+end
+
 local function get_layout(dir, widget1, ...)
     local ret = flex[dir](widget1, ...)
 
-    util.table.crush(ret, ratio, true)
+    gtable.crush(ret, ratio, true)
 
     ret._private.fill_space = nil
 
@@ -376,16 +602,14 @@ function ratio.vertical(...)
     return get_layout("vertical", ...)
 end
 
---Imported documentation
-
-
---- Get a widex index.
+--
+--- Get a widget index.
 -- @param widget The widget to look for
 -- @param[opt] recursive Also check sub-widgets
--- @param[opt] ... Aditional widgets to add at the end of the \"path\"
+-- @param[opt] ... Additional widgets to add at the end of the path
 -- @return The index
 -- @return The parent layout
--- @return The path between \"self\" and \"widget\"
+-- @return The path between self and widget
 -- @function index
 
 --- Get all direct and indirect children widgets.
@@ -545,39 +769,37 @@ end
 -- in its local coordinate system.
 -- @see mouse
 
-
---Imported documentation
-
-
---- Disconnect to a signal.
--- @tparam string name The name of the signal
--- @tparam function func The callback that should be disconnected
+--
+--- Disconnect from a signal.
+-- @tparam string name The name of the signal.
+-- @tparam function func The callback that should be disconnected.
 -- @function disconnect_signal
 
 --- Emit a signal.
 --
--- @tparam string name The name of the signal
+-- @tparam string name The name of the signal.
 -- @param ... Extra arguments for the callback functions. Each connected
---   function receives the object as first argument and then any extra arguments
---   that are given to emit_signal()
+--   function receives the object as first argument and then any extra
+--   arguments that are given to emit_signal().
 -- @function emit_signal
 
 --- Connect to a signal.
--- @tparam string name The name of the signal
--- @tparam function func The callback to call when the signal is emitted
+-- @tparam string name The name of the signal.
+-- @tparam function func The callback to call when the signal is emitted.
 -- @function connect_signal
 
---- Connect to a signal weakly. This allows the callback function to be garbage
--- collected and automatically disconnects the signal when that happens.
+--- Connect to a signal weakly.
+--
+-- This allows the callback function to be garbage collected and
+-- automatically disconnects the signal when that happens.
 --
 -- **Warning:**
 -- Only use this function if you really, really, really know what you
 -- are doing.
--- @tparam string name The name of the signal
--- @tparam function func The callback to call when the signal is emitted
+-- @tparam string name The name of the signal.
+-- @tparam function func The callback to call when the signal is emitted.
 -- @function weak_connect_signal
 
-
 return ratio
 
 -- vim: filetype=lua:expandtab:shiftwidth=4:tabstop=8:softtabstop=4:textwidth=80
